"""
Cost functions of Machine Learning
"""

import numpy as np


def calculate_mse_cost(y_pred: np.ndarray, y: np.ndarray) -> float:
    """Calculate error for regression model with mean squared error

    Args:
        y_pred (np.ndarray): predicted y value, y^.
        y (np.ndarray): actual y value. 

    Returns:
        float: mean squared error cost.
    """
    residual = y_pred - y
    diff_squared = np.square(residual)
    mse_cost = np.mean(diff_squared) / 2
    return float(mse_cost)


def calculate_entropy_cost(y_pred: np.ndarray, y: np.ndarray) -> float:
    """Calculate entropy error for classification model

    Args:
        y_pred (np.ndarray): predicted y value, y^.
        y (np.ndarray): actual y value.

    Returns:
        float: entorpy error cost.
    """

    part_1 = y * np.log(y_pred)

    part_2 = (1 - y) * np.log(1 - y_pred)

    cost = (-1 / y_pred.shape[0]) * np.sum(part_1 + part_2)
    return cost




from mightypy.ml import DecisionTreeClassifier,DecisionTreeRegressor
from sklearn.datasets import load_iris,load_boston
from sklearn.model_selection import train_test_split
from sklearn.metrics import mean_squared_error
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np


dataset = load_iris()

X = dataset.data
y = dataset.target
feature_name = dataset.feature_names
X_train,X_test,y_train,y_test = train_test_split(X,y,random_state=42,test_size=0.10)

dt = DecisionTreeClassifier(criteria='gini')
dt.train(X=X_train,y=y_train,feature_name=feature_name)
y_pred = dt.predict_probability(X_test)
print(pd.DataFrame(np.concatenate((y_test.reshape(-1,1),y_pred),axis=1),columns=['original','predicted']))


dt = DecisionTreeClassifier(criteria='entropy')
dt.train(X=X_train,y=y_train,feature_name=feature_name)
y_pred = dt.predict_probability(X_test)
print(pd.DataFrame(np.concatenate((y_test.reshape(-1,1),y_pred),axis=1),columns=['original','predicted']))


boston_dataset = load_boston()

dataset = pd.DataFrame(data=boston_dataset['data'],columns=boston_dataset.feature_names)
dataset['target'] = boston_dataset["target"]

X = dataset[["RM","AGE","DIS","LSTAT"]].values
y = boston_dataset.target.reshape(-1,1)

feature_name = ["RM","AGE","DIS","LSTAT"]
X_train,X_test,y_train,y_test = train_test_split(X,y,random_state=42,test_size=0.10)

dt = DecisionTreeRegressor(max_depth=100,min_samples_split=3)
dt.train(X=X_train,y=y_train,feature_name=list(feature_name))
y_pred = dt.predict(X_test)
df = pd.DataFrame(np.concatenate((y_test.reshape(-1,1),y_pred),axis=1),columns=['original','predicted'])

plt.plot(df['original'])
plt.plot(df['predicted'])
plt.show()


mean_squared_error(df['original'],df['predicted'])



"""
data drift module
==================

"""
from typing import Union
import numpy as np
import pandas as pd


def population_stability_index(expected: Union[list, np.ndarray], actual: Union[list, np.ndarray], data_type: str) -> pd.DataFrame:
    """
    Populaion Stability Index.

    References:
        https://www.listendata.com/2015/05/population-stability-index.html

    Args:
        expected (Union[list, np.ndarray]): Expected values.
        actual (Union[list, np.ndarray]): Actual values.
        data_type (str): Type of data. Helps in bucketing.

    Returns:
        pd.DataFrame: calculated dataframe.

    Examples:
        >>> import numpy as np
        >>> from mightypy.stats import population_stability_index
        
        continuous data
            >>> expected_continuous = np.random.normal(size=(500,))
            >>> actual_continuous = np.random.normal(size=(500,))
            >>> psi_df = population_stability_index(expected_continuous, actual_continuous, data_type='continuous')
            >>> psi_df.psi.sum()

        discrete data
            >>> expected_discrete = np.random.randint(0,10, size=(500,))
            >>> actual_discrete = np.random.randint(0,10, size=(500,))
            >>> psi_df = population_stability_index(expected_discrete, actual_discrete, data_type='discrete')
            >>> psi_df.psi.sum()
    """
    if data_type == 'continuous':
        max_val: Union[int, float] = np.max(expected)
        min_val: Union[int, float] = np.min(expected)

        ranges = np.linspace(min_val, max_val, 11)[1:-1]
        bins = [-np.inf, *ranges, np.inf]
        labels = [
            f"{idx+1} | {i[0]:.2f} to {i[1]:.2f}" for idx, i in enumerate(zip(bins[:-1], bins[1:]))
        ]
        expected_cuts = pd.cut(expected, bins=bins, labels=labels).value_counts()
        actual_cuts = pd.cut(actual, bins=bins, labels=labels).value_counts()

    elif data_type == 'discrete':
        expected_cuts = pd.Series(expected).value_counts()
        actual_cuts = pd.Series(actual).value_counts()

    else:
        raise NotImplementedError(f"Method {data_type} is not implemented, or correct one. Try continuous, discrete.")

    calc_df = pd.concat([expected_cuts, actual_cuts], axis=1,
                        keys=['expected', 'actual']).sort_index()
    calc_df[['expected %', 'actual %']] = (
        calc_df[['expected', 'actual']]/calc_df[['expected', 'actual']].sum(axis=0))
    calc_df['diff'] = calc_df['actual %'] - calc_df['expected %']
    calc_df['log(actual %/ expected %)'] = np.log(calc_df['actual %'] /
                                                  calc_df['expected %'])
    calc_df['psi'] = calc_df['diff'] * calc_df['log(actual %/ expected %)']
    return calc_df


if __name__ == "__main__":

    expected_continuous = np.random.normal(size=(500,))
    actual_continuous = np.random.normal(size=(500,))

    expected_discrete = np.random.randint(0,10, size=(500,))
    actual_discrete = np.random.randint(0,10, size=(500,))

    print(population_stability_index(expected_continuous, actual_continuous, data_type='continuous'))

    print(population_stability_index(expected_discrete, actual_discrete, data_type='discrete'))



from __future__ import annotations
from typing import Union, Tuple, List
import warnings
import numpy as np


class Question:
    """Question is a thershold/matching concept for splitting the node of the Decision Tree

    Args:
        column_index (int): Column index to be chosen from the array passed at the matching time.
        value (Union[int, str, float, np.int64, np.float64]): Threshold value/ matching value.
        header (str): column/header name.
    """

    def __init__(self, column_index: int, value: Union[int, str, float, np.int64, np.float64], header: str):
        """Constructor
        """
        self.column_index = column_index
        self.value = value
        self.header = header

    def match(self, example: Union[list, np.ndarray]) -> bool:
        """Matching function to decide based on example whether result is true or false.

        Args:
            example (Union[list, np.ndarray]): Example to compare with question parameters.
        Returns:
            bool: if the example is in threshold or value matches then results true or false.
        """
        if isinstance(example, list):
            example = np.array(example, dtype="O")
        val = example[self.column_index]

        # adding numpy int and float data types as well
        if isinstance(val, (int, float, np.int64, np.float64)):
            # a condition for question to return True or False for numeric value
            return float(val) >= float(self.value)
        else:
            return str(val) == str(self.value)  # categorical data comparison

    def __repr__(self):
        condition = "=="
        if isinstance(self.value, (int, float, np.int64, np.float64)):
            condition = ">="
        return f"Is {self.header} {condition} {self.value} ?"


class Node:
    """A Tree node either Decision Node or Leaf Node

    Args:
        question (Question, optional): question object. Defaults to None.
        true_branch (Node, optional): connection to node at true side of the branch. Defaults to None.
        false_branch (Node, optional): connection to node at false side of the branch. Defaults to None.
        uncertainty (float, optional): Uncertainty value like gini,entropy,variance etc. Defaults to None.
        leaf_value (Union[dict,int,float], optional): Leaf node/final node's value. Defaults to None.
    """

    def __init__(self, question: Question = None, true_branch: Node = None, false_branch: Node = None, uncertainty: float = None, *, leaf_value: Union[dict, int, float] = None):
        """Constructor
        """
        self.question = question
        self.true_branch = true_branch
        self.false_branch = false_branch
        self.uncertainty = uncertainty
        self.leaf_value = leaf_value

    @property
    def _is_leaf_node(self) -> bool:
        """Check if this node is leaf node or not.
        Returns:
            bool: True if leaf node else false.
        """
        return self.leaf_value is not None


class DecisionTreeClassifier:
    """Decision Tree Based Classification Model

    Args:
        max_depth (int, optional): max depth of the tree. Defaults to 100.
        min_samples_split (int, optional): min size of the sample at the time of split. Defaults to 2.
        criteria (str, optional): what criteria to use for information. Defaults to 'gini'. available 'gini','entropy'.
    """

    def __init__(self, max_depth: int = 100, min_samples_split: int = 2, criteria: str = 'gini'):
        """Constructor
        """
        self._X = None
        self._y = None
        self._feature_names = None
        self._target_name = None
        self._tree = None
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.criteria = criteria

    def _count_dict(self, a: np.ndarray) -> dict:
        """Count class frequecies and get a dictionary from it

        Args:
            a (np.ndarray): input array. shape should be (m,1) for m samples.

        Returns:
            dict: categories/classes freq dictionary.
        """
        unique_values = np.unique(a, return_counts=True)
        zipped = zip(*unique_values)
        dict_obj = dict(zipped)
        return dict_obj

    def _gini_impurity(self, arr: np.ndarray) -> float:
        """Calculate Gini Impurity

        Args:
            arr (np.ndarray): input array.

        Returns:
            float: gini impurity value.
        """
        classes, counts = np.unique(arr, return_counts=True)
        gini_score = 1 - np.square(counts / arr.shape[0]).sum(axis=0)
        return gini_score

    def _entropy(self, arr: np.ndarray) -> float:
        """Calculate Entropy

        Args:
            arr (np.ndarray): input array.

        Returns:
            float: entropy result.
        """
        classes, counts = np.unique(arr, return_counts=True)
        p = counts / arr.shape[0]
        entropy_score = (-p * np.log2(p)).sum(axis=0)
        return entropy_score

    def _uncertainty(self, a: np.ndarray) -> float:
        """calcualte uncertainty

        Args:
            a (np.ndarray): input array

        Returns:
            float: uncertainty value
        """

        if self.criteria == "entropy":
            value = self._entropy(a)
        elif self.criteria == "gini":
            value = self._gini_impurity(a)
        else:
            warnings.warn(f"{self.criteria} is not coded yet. returning to gini.")
            value = self._gini_impurity(a)
        return value

    def _partition(self, rows: np.ndarray, question: Union[Question, None]) -> Tuple[list, list]:
        """partition the rows based on the question

        Args:
            rows (np.ndarray): input array to split.
            question (Question): question object containing spltting concept.

        Returns:
            Tuple[list,list]: true idxs and false idxs.
        """
        true_idx, false_idx = [], []
        for idx, row in enumerate(rows):
            if question.match(row):
                true_idx.append(idx)
            else:
                false_idx.append(idx)
        return true_idx, false_idx

    def _info_gain(self, left: np.ndarray, right: np.ndarray, parent_uncertainty: float) -> float:
        """Calculate information gain after splitting

        Args:
            left (np.ndarray): left side array.
            right (np.ndarray): right side array.
            parent_uncertainty (float): parent node Uncertainity.

        Returns:
            flaot: information gain value.
        """

        # calculating portion/ partition/ weightage
        pr = left.shape[0] / (left.shape[0] + right.shape[0])

        # calcualte child uncertainity
        child_uncertainty = pr * \
            self._uncertainty(left) - (1 - pr) * self._uncertainty(right)
        # calculate information gain
        info_gain_value = parent_uncertainty - child_uncertainty
        return info_gain_value

    def _find_best_split(self, X: np.ndarray, y: np.ndarray) -> Tuple[float, Union[Question, None], float]:
        """method to find best split possible for the sample
        
        Args:
            X (np.ndarray): Feature matrix.
            y (np.ndarray): target matrix.
        
        Returns:
            Tuple[float,Union[Question,None],float]: maximum gain from the split, best question of it, and parent node uncertainty.
        """

        max_gain = -1
        best_split_question = None

        parent_uncertainty = self._uncertainty(y)

        m_samples, n_labels = X.shape

        for col_index in range(n_labels):  # iterate over feature columns
            # get unique values from the feature
            unique_values = np.unique(X[:, col_index])
            for val in unique_values:  # check for every value and find maximum info gain

                ques = Question(
                    column_index=col_index,
                    value=val,
                    header=self._feature_names[col_index]
                )

                t_idx, f_idx = self._partition(X, ques)
                # if it does not split the data
                # skip it
                if len(t_idx) == 0 or len(f_idx) == 0:
                    continue

                true_y = y[t_idx, :]
                false_y = y[f_idx, :]

                # get information gain
                gain = self._info_gain(true_y, false_y, parent_uncertainty)
                if gain > max_gain:
                    max_gain, best_split_question = gain, ques

        return max_gain, best_split_question, parent_uncertainty

    def _build_tree(self, X: np.ndarray, y: np.ndarray, depth: int = 0) -> Node:
        """Recursive funtion to build tree.

        Args:
            X (np.ndarray): input features matrix.   
            y (np.ndarray): target matrix.
            depth (int, optional): depth count of the recursion. Defaults to 0.

        Returns:
            Node: either leaf node or decision node
        """
        m_samples, n_labels = X.shape

        # if depth is greater than max depth defined or labels/features are left to 1
        # or number of samples are less than the minimum size of samples to split then
        # stop recursion and return a node
        if (depth > self.max_depth or n_labels == 1 or m_samples < self.min_samples_split):
            return Node(leaf_value=self._count_dict(y))

        gain, ques, uncertainty = self._find_best_split(X, y)

        # if gain is zero
        # then no point grinding further here
        if gain < 0:
            return Node(leaf_value=self._count_dict(y))

        t_idx, f_idx = self._partition(X, ques)  # get partition idxs
        true_branch = self._build_tree(
            X[t_idx, :], y[t_idx, :], depth + 1)  # recog true branch samples
        false_branch = self._build_tree(
            X[f_idx, :], y[f_idx, :], depth + 1)  # recog false branch samples
        return Node(
            question=ques,
            true_branch=true_branch,
            false_branch=false_branch,
            uncertainty=uncertainty
        )

    def train(self, X: Union[np.ndarray, list], y: Union[np.ndarray, list], feature_name: list = None, target_name: list = None) -> None:
        """Train the model

        Args:
            X (Union[np.ndarray,list]): feature matrix.
            y (Union[np.ndarray,list]): target matrix.
            feature_name (list, optional): feature names list. Defaults to None.
            target_name (list, optional): target name list. Defaults to None.
        """

        X = np.array(X, dtype='O') if not isinstance(
            X, (np.ndarray)) else X  # converting to numpy array
        y = np.array(y, dtype='O') if not isinstance(
            y, (np.ndarray)) else y  # converting to numpy array

        # reshaping to vectors
        self._X = X.reshape(-1, 1) if len(X.shape) == 1 else X
        self._y = y.reshape(-1, 1) if len(y.shape) == 1 else y

        # creating feature names if not mentioned
        self._feature_names = feature_name or [
            f"C_{i}" for i in range(self._X.shape[1])]

        # creating target name if not mentioned
        self._target_name = target_name or ['target']

        # BOOOM
        # building the tree
        self._tree = self._build_tree(
            X=self._X,
            y=self._y
        )

    def print_tree(self, node: Union[Node, None] = None, spacing: str = "|-") -> None:
        """print the tree

        Args:
            node (Union[Node,None], optional): starting node. Defaults to None. then it will go to the root node of the tree.
            spacing (str, optional): printing separater. Defaults to "|-".
        """

        node = node or self._tree

        if node._is_leaf_node:
            print(spacing, " Predict :", node.leaf_value)
            return

        # Print the question at this node
        print(spacing + str(node.question) +
              " | " + self.criteria + " :" + str(node.uncertainty))

        # Call this function recursively on the true branch
        print(spacing + '--> True:')
        self.print_tree(node.true_branch, "  " + spacing + "-")

        # Call this function recursively on the false branch
        print(spacing + '--> False:')
        self.print_tree(node.false_branch, "  " + spacing + "-")

    def _classification(self, row: np.ndarray, node: Union[Node, None]) -> Union[dict]:
        """Classification recursive function

        Args:
            row (np.ndarray): input matrix.
            node (Union[Node,None]): node to start with. mostly root node. rest will be handled by recursion.

        Returns:
            Union[dict]: leaf value. classification result.
        """
        if node._is_leaf_node:
            return node.leaf_value

        if node.question.match(row):
            return self._classification(row, node.true_branch)
        else:
            return self._classification(row, node.false_branch)

    def _leaf_probabilities(self, results: dict) -> dict:
        """get probabilties

        Args:
            results (dict): results from _classification.

        Returns:
            dict: dictionary with categorical probabilities.
        """
        total = sum(results.values())
        probs = {}
        for key in results:
            probs[key] = (results[key] / total) * 100
        return probs

    def predict(self, X: Union[np.ndarray, list]) -> np.ndarray:
        """predict classification results

        Args:
            X (Union[np.ndarray,list]): testing matrix.

        Raises:
            ValueError: input X can only be a list or numpy array.

        Returns:
            np.ndarray: results of classification.
        """
        if isinstance(X, (np.ndarray, list)):
            X = np.array(X, dtype='O') if not isinstance(X, (np.ndarray)) else X

            if len(X.shape) == 1:
                max_result = 0
                result_dict = self._classification(row=X, node=self._tree)
                result = None
                for key in result_dict:
                    if result_dict[key] > max_result:
                        result = key
                return np.array([[result]], dtype='O')
            else:
                leaf_value = []
                # get maximum caterigorical value from all catergories
                for row in X:
                    max_result = 0
                    result_dict = self._classification(row=row, node=self._tree)
                    result = None
                    for key in result_dict:
                        if result_dict[key] > max_result:
                            result = key
                    leaf_value.append([result])
                return np.array(leaf_value, dtype='O')
        else:
            raise ValueError("X should be list or numpy array")

    def predict_probability(self, X: Union[np.ndarray, list]) -> Union[np.ndarray, dict]:
        """predict classfication probabilities

        Args:
            X (Union[np.ndarray,list]): testing matrix.

        Raises:
            ValueError: input X can only be a list or numpy array.

        Returns:
            Union[np.ndarray, dict]: probabity results of classification.
        """
        if isinstance(X, (np.ndarray, list)):
            X = np.array(X, dtype='O') if not isinstance(X, (np.ndarray)) else X

            if len(X.shape) == 1:
                return self._leaf_probabilities(self._classification(row=X, node=self._tree))
            else:
                leaf_value = []
                for row in X:
                    leaf_value.append([self._leaf_probabilities(
                        self._classification(row=row, node=self._tree))])
                return np.array(leaf_value, dtype='O')
        else:
            raise ValueError("X should be list or numpy array")


class DecisionTreeRegressor:
    """Decision Tree Based Regression Model

    Args:
        max_depth (int, optional): maximum depth of the tree. Defaults to 10.
        min_samples_split (int, optional): minimum number of samples while splitting. Defaults to 3.
        criteria (str, optional): criteria for best info gain. Defaults to 'variance'.
    """

    def __init__(self, max_depth: int = 10, min_samples_split: int = 3, criteria: str = 'variance'):
        """constructor
        """
        self._X = None
        self._y = None
        self._feature_names = None
        self._target_name = None
        self._tree = None
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.criteria = criteria

    def _mean_leaf_value(self, a: np.ndarray) -> float:
        """leaf values mean

        Args:
            a (np.ndarray): input array.

        Returns:
            float: mean value
        """

        return float(np.mean(a))

    def _partition(self, rows: np.ndarray, question: Union[Question, None]) -> Tuple[list, list]:
        """partition the rows based on the question

        Args:
            rows (np.ndarray): input array to split.
            question (Question): question object containing spltting concept.

        Returns:
            Tuple[list,list]: true idxs and false idxs.
        """
        true_idx, false_idx = [], []
        for idx, row in enumerate(rows):
            if question.match(row):
                true_idx.append(idx)
            else:
                false_idx.append(idx)
        return true_idx, false_idx

    def _uncertainty(self, a: np.ndarray) -> float:
        """calcualte uncertainty

        Args:
            a (np.ndarray): input array

        Returns:
            float: uncertainty value
        """
        if self.criteria == "variance":
            value = np.var(a)
        else:
            warnings.warn(f"{self.criteria} is not coded yet. returning to variance.")
            value = np.var(a)
        return float(value)

    def _info_gain(self, left: np.ndarray, right: np.ndarray, parent_uncertainty: float) -> float:
        """Calculate information gain after splitting

        Args:
            left (np.ndarray): left side array.
            right (np.ndarray): right side array.
            parent_uncertainty (float): parent node Uncertainity.

        Returns:
            flaot: information gain value.
        """

        pr = left.shape[0] / (left.shape[0] + right.shape[0])

        child_uncertainty = pr * \
            self._uncertainty(left) - (1 - pr) * self._uncertainty(right)
        info_gain_value = parent_uncertainty - child_uncertainty
        return info_gain_value

    def _find_best_split(self, X: np.ndarray, y: np.ndarray) -> Tuple[float, Union[Question, None], float]:
        """method to find best split possible for the sample

        Args:
            X (np.ndarray): Feature matrix.
            y (np.ndarray): target matrix.

        Returns:
            Tuple[float,Union[Question,None],float]: maximum gain from the split, best question of it, and parent node uncertainty
        """

        max_gain = -1
        best_split_question = None

        parent_uncertainty = self._uncertainty(y)

        m_samples, n_labels = X.shape

        for col_index in range(n_labels):  # iterate over feature columns
            # get unique values from the feature
            unique_values = np.unique(X[:, col_index])
            for val in unique_values:  # check for every value and find maximum info gain

                ques = Question(
                    column_index=col_index,
                    value=val,
                    header=self._feature_names[col_index]
                )

                t_idx, f_idx = self._partition(X, ques)
                # if it does not split the data
                # skip it
                if len(t_idx) == 0 or len(f_idx) == 0:
                    continue

                true_y = y[t_idx, :]
                false_y = y[f_idx, :]

                gain = self._info_gain(true_y, false_y, parent_uncertainty)
                if gain > max_gain:
                    max_gain, best_split_question = gain, ques

        return max_gain, best_split_question, parent_uncertainty

    def _build_tree(self, X: np.ndarray, y: np.ndarray, depth: int = 0) -> Node:
        """Recursive funtion to build tree

        Args:
            X (np.ndarray): input features matrix.   
            y (np.ndarray): target matrix.
            depth (int, optional): depth count of the recursion. Defaults to 0.

        Returns:
            Node: either leaf node or decision node
        """
        m_samples, n_labels = X.shape

        # if depth is greater than max depth defined or labels/features are left to 1
        # or number of samples are less than the minimum size of samples to split then
        # stop recursion and return a node
        if (depth > self.max_depth or n_labels == 1 or m_samples < self.min_samples_split):
            return Node(leaf_value=y)

        gain, ques, uncertainty = self._find_best_split(X, y)

        # if gain is zero no point in going further
        if gain < 0:
            return Node(leaf_value=y)

        t_idx, f_idx = self._partition(X, ques)
        true_branch = self._build_tree(
            X[t_idx, :], y[t_idx, :], depth + 1)  # get true samples
        false_branch = self._build_tree(
            X[f_idx, :], y[f_idx, :], depth + 1)  # get false samples
        return Node(
            question=ques,
            true_branch=true_branch,
            false_branch=false_branch,
            uncertainty=uncertainty
        )

    def train(self, X: Union[np.ndarray, list], y: Union[np.ndarray, list], feature_name: list = None, target_name: list = None) -> None:
        """Train the model

        Args:
            X (Union[np.ndarray,list]): feature matrix.
            y (Union[np.ndarray,list]): target matrix.
            feature_name (list, optional): feature names list. Defaults to None.
            target_name (list, optional): target name list. Defaults to None.
        """

        X = np.array(X, dtype='O') if not isinstance(
            X, (np.ndarray)) else X  # converting to numpy array
        y = np.array(y, dtype='O') if not isinstance(
            y, (np.ndarray)) else y  # converting to numpy array

        # reshaping to vectors
        self._X = X.reshape(-1, 1) if len(X.shape) == 1 else X
        self._y = y.reshape(-1, 1) if len(y.shape) == 1 else y

        # creating feature names if not mentioned
        self._feature_names = feature_name or [
            f"C_{i}" for i in range(self._X.shape[1])]

        # creating target name if not mentioned
        self._target_name = target_name or ['target']

        # BOOOM
        # building the tree
        self._tree = self._build_tree(
            X=self._X,
            y=self._y
        )

    def print_tree(self, node: Union[Node, None] = None, spacing: str = "|-", mean_preds: bool = True) -> None:
        """print the tree

        Args:
            node (Union[Node,None], optional): starting node. Defaults to None. then it will go to the root node of the tree.
            spacing (str, optional): printing separater. Defaults to "|-".
            mean_preds (bool): do the mean of prediction values. Defaults to True.
        """

        node = node or self._tree

        if node._is_leaf_node:
            if mean_preds:
                print(spacing, " Predict :", self._mean_leaf_value(node.leaf_value))
            else:
                print(spacing, " Predict :", node.leaf_value[...,-1])
            return

        # Print the question at this node
        print(spacing + str(node.question) +
              " | " + self.criteria + " :" + str(node.uncertainty))

        # Call this function recursively on the true branch
        print(spacing + '--> True:')
        self.print_tree(node.true_branch, "  " + spacing + "-", mean_preds)

        # Call this function recursively on the false branch
        print(spacing + '--> False:')
        self.print_tree(node.false_branch, "  " + spacing + "-", mean_preds)

    def _regression(self, row: np.ndarray, node: Union[Node, None], mean_preds: bool) -> float:
        """regression recursive method

        Args:
            row (np.ndarray): input matrix.
            node (Union[Node,None]): node to start with. mostly root node. rest will be handled by recursion.
            mean_preds (bool): do the mean of prediction values.
            
        Returns:
            float: regression result.
        """

        if node._is_leaf_node:
            if mean_preds:
                return self._mean_leaf_value(node.leaf_value)
            else:
                return node.leaf_value[...,-1]

        if node.question.match(row):
            return self._regression(row, node.true_branch, mean_preds)
        else:
            return self._regression(row, node.false_branch, mean_preds)

    def predict(self, X: np.ndarray, mean_preds: bool = True) -> np.ndarray:
        """predict regresssion
        Args:
            X (np.ndarray): testing matrix.
            mean_preds (bool): do the mean of prediction values. Defaults to True.
        Raises:
            ValueError: X should be list or numpy array
        Returns:
            np.ndarray: regression prediction.
        """
        if isinstance(X, (np.ndarray, list)):
            X = np.array(X, dtype='O') if not isinstance(X, (np.ndarray)) else X

            if len(X.shape) == 1:
                result = self._regression(row=X, node=self._tree, mean_preds=mean_preds)
                return np.array([[result]], dtype='O')
            else:
                leaf_value = []
                for row in X:
                    result = self._regression(row=row, node=self._tree, mean_preds=mean_preds)
                    leaf_value.append([result])
                return np.array(leaf_value, dtype='O')
        else:
            raise ValueError("X should be list or numpy array")





""" 
mightypy.dsa
==============
"""

from ._linked_list import (
    LinkedList
)

from ._search import (
    binary_search,
    linear_search
)

from ._tree import (
    BinaryTree,
    BST
)


__all__ = [
    "LinkedList",
    "binary_search",
    "linear_search",
    "BinaryTree",
    "BST"
]


"""
mightypy.ml
=============
"""

from ._linear import (
    LinearRegression,
    LassoRegression,
    RidgeRegression,
    LogisticRegression,
    polynomial_regression,
    trend
)


from ._tree import (
    DecisionTreeClassifier,
    DecisionTreeRegressor
)

from ._ensemble import (
    AdaboostClassifier,
    RandomForestClassifier,
    RandomForestRegressor,
)

from ._utils import (
    moving_window_matrix,
    sigmoid
)


__all__ = [
    "LinearRegression",
    "LassoRegression",
    "RidgeRegression",
    "LogisticRegression",
    "polynomial_regression",
    "trend",
    "DecisionTreeClassifier",
    "DecisionTreeRegressor",
    "AdaboostClassifier",
    "RandomForestClassifier",
    "RandomForestRegressor",
    "moving_window_matrix",
    "sigmoid"
]



"""
mightypy.make
===============
"""


from ._dataset import (
    spiral_data,
    rotation_matrix_2d,
    sine_wave_from_sample,
    sine_wave_from_timesteps
)


__all__ = [
    "spiral_data",
    "rotation_matrix_2d",
    "sine_wave_from_sample",
    "sine_wave_from_timesteps"
]


"""
searching methods
"""

from typing import Union, Tuple, List


def linear_search(arr: List[Union[int, float, str]], ele: Union[int, float, str]) -> Union[Tuple[bool, int], Tuple[bool, None]]:
    """Linear search algorithm

    Args:
        arr (List[Union[int,float,str]]): array list to search
        ele (Union[int, float, str]): element to search

    Returns:
        Union[Tuple[bool, int], Tuple[bool, None]]: result of linear search
    """
    for i in range(len(arr)):
        if arr[i] == ele:
            return True, i

    return False, None


def binary_search(arr: List[Union[int, float, str]], ele: Union[int, float, str]) -> Union[Tuple[bool, int], Tuple[bool, None]]:
    """Binary search algorithm

    Args:
        arr (List[Union[int,float,str]]): array list to search
        ele (Union[int, float, str]): element to search

    Returns:
        Union[Tuple[bool, int], Tuple[bool, None]]: result of bianry search
    """
    low: int = 0
    high: int = len(arr) - 1

    while low <= high:
        mid = int(low + (high - low) / 2)
        if arr[mid] == ele:
            return True, mid
            
        if arr[mid] < ele:
            low = mid + 1
        else:
            high = mid - 1
    return False, None


"""
Binary tree setup
"""

class Node:
    """A node for binary tree
    """

    def __init__(self, data):
        self.data = data
        self._left = None
        self._right = None

    @property
    def left(self):
        return self._left
    
    @left.setter
    def left(self,new_left):
        self._left = new_left

    @property
    def right(self):
        return self._right
    
    @right.setter
    def right(self,new_right):
        self._right = new_right

    def __repr__(self) -> str:
        return f"""Tree Node Value:{self.data}"""



class BinaryTree:

    def __init__(self) -> None:
        self._root = None   

    @staticmethod
    def _calc_height(node):
        if node is None:
            return 0
        else:
            # Compute the height of each subtree
            lheight = BinaryTree._calc_height(node.left)
            rheight = BinaryTree._calc_height(node.right)
    
            # Use the larger one
            return max(lheight, rheight) + 1

    @property
    def height(self):
        return self._calc_height(self._root)

    def insert(self,val):
        """
        Level order insertion.

        https://en.wikipedia.org/wiki/Breadth-first_search

        Args:
            val (Any): value of the node in tree.
        """
        if not self._root:
            self._root = Node(val)
            return

        q = []
        q.append(self._root)
        while len(q):
            temp = q.pop(0)

            if not temp.left:
                temp.left = Node(val)
                break
            else:
                q.append(temp.left)

            if not temp.right:
                temp.right = Node(val)
                break
            else:
                q.append(temp.right)

    def _level_order_traverse(self):
        if self._root is None:
            return
        
        q = []
        values = []
        q.append(self._root)
        while len(q):
            values.append(q[0].data)
            node = q.pop(0)

            if node.left is not None:
                q.append(node.left)
            
            if node.right is not None:
                q.append(node.right)
        return values

    def _inorder_traverse_rec(self, node, values=[]):
        """Inorder traversal recursive function

        Args:
            node (Node): tree node
            values (list, optional): list of values. Defaults to [].

        Returns:
            values (list): list of values
        """
        if node:
            data = node.data
            left = node.left
            right = node.right

            self._inorder_traverse_rec(left, values)
            values.append(data)
            self._inorder_traverse_rec(right, values)
        return values

    def _preorder_traverse_rec(self, node, values=[]):
        """preorder traversal recursive function

        Args:
            node (Node): tree node
            values (list, optional): list of values. Defaults to [].

        Returns:
            values (list): list of values
        """
        if node:
            data = node.data
            left = node.left
            right = node.right

            values.append(data)
            self._preorder_traverse_rec(left, values)
            self._preorder_traverse_rec(right, values)
        return values

    def _postorder_traverse_rec(self, node, values=[]):
        """Postorder traversal recursive function

        Args:
            node (Node): tree node
            values (list, optional): list of values. Defaults to [].

        Returns:
            values (list): list of values
        """
        if node:
            data = node.data
            left = node.left
            right = node.right

            self._postorder_traverse_rec(left, values)
            self._postorder_traverse_rec(right, values)
            values.append(data)
        return values

    def _inorder_traverse_stack(self):
        """Inorder traversal using stack

        Notes:
            Algorithm 

                1. Create an empty stack S.
                2. Initialize current node as root
                3. Push the current node to S and set current = current->left until current is NULL
                4. If current is NULL and stack is not empty then

                    4.1. Pop the top item from stack
                    4.2. Print the popped item, set current = popped_item->right
                    4.3. Go to step 3
                    
                5. If current is NULL and stack is empty then we are done.

        Returns:
            node_values (lsit): tree values
        """

        node_values = []
        current_node = self._root
        stack = []

        while True:
            if current_node is not None:
                stack.append(current_node)
                current_node = current_node.left

            elif stack:  # if stack is empty then true
                current_node = stack.pop()  # pop the node top of the stack
                node_values.append(current_node.data)
                current_node = current_node.right

            else:
                break
        return node_values

    def _preorder_traverse_stack(self):
        """Preorder Traversal using Stack

        Notes:
            Algorithm
                TODO

        Returns:
            node_values (list): node values
        """
        node_values = []
        current_node = self._root
        stack = []

        while True:
            if current_node is not None:
                node_values.append(current_node.data)
                stack.append(current_node)
                current_node = current_node.left

            elif stack:  # if stack is empty then true
                current_node = stack.pop()  # pop the node top of the stack
                current_node = current_node.right
            else:
                break
        return node_values

    def _postorder_traverse_stack(self):
        """Postorder Traversal using Stack

        Notes:
            Algorithm
                TODO
                
        Returns:
            node_values (list): node values
        """
        node_values = []
        current_node = self._root
        stack = []

        while True:
            if current_node is not None:
                stack.append(current_node)
                stack.append(current_node)
                current_node = current_node.left

            elif stack:  # if stack is empty then true
                current_node = stack.pop()  # pop the node top of the stack

                if len(stack) > 0 and stack[-1] == current_node:
                    current_node = current_node.right
                else:
                    node_values.append(current_node.data)
                    current_node = None
            else:
                break
        return node_values

    def traverse(self, order="in", method="stack"):
        """Tree traversal operation

        Args:
            order (str, optional): order in which the tree will be traversed. Defaults to "in".
                                    Options available -
                                    "level","in","pre","post"

        Raises:
            ValueError: If wrong order is passed. only "level","in","pre","post" is allowed

        Returns:
            values (list): Values of tree nodes in specified order
        """
        order = order.lower()
        method = method.lower()

        if order=="level":
            values = self._level_order_traverse()
        else:
            if method == "recursion":
                if order == "in":
                    values = self._inorder_traverse_rec(node=self._root, values=[])
                elif order == "pre":
                    values = self._preorder_traverse_rec(node=self._root, values=[])
                elif order == "post":
                    values = self._postorder_traverse_rec(node=self._root, values=[])
                else:
                    raise ValueError(f"order : '{order}' is not defined.")

            elif method == "stack":
                if order == "in":
                    values = self._inorder_traverse_stack()
                elif order == "pre":
                    values = self._preorder_traverse_stack()
                elif order == "post":
                    values = self._postorder_traverse_stack()
                else:
                    raise ValueError(f"order : '{order}' is not defined.")

            else:
                raise ValueError(f"method : '{method}' is not defined.")

        return values

    @staticmethod
    def _invert(node):

        node.left, node.right = node.right, node.left
        if node.left:
            BinaryTree._invert(node=node.left)
        if node.right:
            BinaryTree._invert(node=node.right)

    def invert(self):
        BinaryTree._invert(node=self._root)


    
class BST:

    def __init__(self) -> None:
        self._root = None

    @staticmethod
    def _insert(node, val):
        if node is None:
            return Node(val)
        else:
            if node.data == val:
                pass
            elif node.data > val:
                node.left = BST._insert(node=node.left,val=val)
            else:
                node.right = BST._insert(node=node.right,val=val)

        return node

    def insert(self, val):
        self._root = BST._insert(self._root,val)

    @staticmethod
    def _inorder_traverse(node, data_arr = []):
        if node is not None:

            BST._inorder_traverse(node.left, data_arr)
            data_arr.append(node.data)
            BST._inorder_traverse(node.right, data_arr)
            
        return data_arr

    @staticmethod
    def _preorder_traverse(node, data_arr = []):
        if node is not None:
            data_arr.append(node.data)
            BST._preorder_traverse(node.left, data_arr)
            BST._preorder_traverse(node.right, data_arr)
            
        return data_arr

    @staticmethod
    def _postorder_traverse(node, data_arr = []):
        if node is not None:
            BST._postorder_traverse(node.left, data_arr)
            BST._postorder_traverse(node.right, data_arr)
            data_arr.append(node.data)
            
        return data_arr

    

    @staticmethod
    def _level_order_traverse(node):
        if node is None:
            return []
        else:
            # actually this queue is working like a temp storage
            # this temp storage will store nodes that needs to be processed in FIFO
            # as we are talking about level order
            # so it needs to go by this order
            #                    
            #                         root
            #                      /       \
            #                 left1         right1
            #                /    \          /    \
            #             left2   right2   left3   right3
            # root -> left1 -> right1 -> left2 -> right2 -> left3 -> right3
            # 
            # queue's condition
            # [] <- root
            # [root]
            # root <-[ left1 right1 ]
            # left1 <- [ right1 left2 right2 ]
            # right1 <- [ left2 right2 left3 right3 ]
            # left2 <- [ right2 left3 right3 ]
            # right2 <- [ left3 right3 ]
            # left3 <- [ right3 ]
            # right3 <- []
        
            queue = []
            queue.append(node)

            # we need one data array to store traversed data no
            data_arr = []

            # now run the loop until we reach to an empty queue
            while len(queue):

                # take the first element from the queue and add it in data_arr
                # remove the element from queue but use it for left and right nodes iteration
                curr_node = queue.pop(0)

                data_arr.append(curr_node.data)
                
                if curr_node.left is not None: # adding left node to queue
                    queue.append(curr_node.left)
                
                if curr_node.right is not None: # adding right node to queue
                    queue.append(curr_node.right)
            
                # based on FIFO queue will process left node first 
                # then the right node.
            return data_arr


    def traverse(self, order="level"):
        
        order = order.lower()

        if order == "level":
            return BST._level_order_traverse(node=self._root)
        elif order == "in":
            return BST._inorder_traverse(node=self._root)
        elif order == "pre":
            return BST._preorder_traverse(node=self._root)
        elif order == "post":
            return BST._postorder_traverse(node=self._root)
        else:
            pass

if __name__ == "__main__":

    # tree = BinaryTree()
    # for i in range(1,7):
    #     tree.insert(i)


    # print("height", tree.height)

    # print("level", tree.traverse(order="level", method="stack"))

    # tree.invert()

    # print("level", tree.traverse(order="level", method="stack"))
    # print("height", tree.height)


    # print("post", tree.traverse(order="post",method="stack"))
    # print("post", tree.traverse(order="post", method="recursion"))

    # print("pre", tree.traverse(order="pre",method="stack"))
    # print("pre", tree.traverse(order="pre", method="recursion"))

    # print("in", tree.traverse(order="in", method="stack"))
    # print("in", tree.traverse(order="in", method="recursion"))


    bst = BST()

    for i in [3,4,6,12,9,11,5]:
        bst.insert(val=i)

    print(bst.traverse(order="in"))
    



"""
Linear Models of Machine Learning
"""

from typing import Union, Tuple
import numpy as np
from mightypy import __version__
from mightypy.ml._error_fxs import calculate_mse_cost, calculate_entropy_cost
from mightypy.ml._utils import sigmoid

__author__ = "Nishant Baheti"
__copyright__ = "Nishant Baheti"
__license__ = "MIT"

class LinearRegression:
    """Linear Regression Model Class

    Args:
        alpha (float, optional): learning rate. Defaults to 0.01.
        iterations (int, optional): number of iteratons. Defaults to 10000.

    References:
        https://machinelearningexploration.readthedocs.io/en/latest/LinearRegression/Explore.html
    """

    def __init__(self, alpha: float = 0.01, iterations: int = 10000):
        """Constructor
        """
        self.alpha = alpha
        self.iterations = iterations
        self._theta = None
        self._X = None
        self._y = None
        self._theta_history = None
        self._cost_history = None
        self._n = None
        self._m = None

    def _format_x_for_theta_0(self, x_i: np.ndarray) -> np.ndarray:
        """format X matrix for linear model

        put 1's in the first column as feature for theta_0

        Args:
            x_i (np.ndarray): input x matrix

        Returns:
            np.ndarray: formatted x matrix
        """
        x_i = x_i.copy()
        if len(x_i.shape) == 1:
            x_i = x_i.reshape(-1, 1)

        if False in (x_i[..., 0] == 1):
            return np.hstack(tup=(np.ones(shape=(x_i.shape[0], 1)), x_i))
        else:
            return x_i

    @property
    def X(self) -> Union[np.ndarray, None]:
        """property X

        Returns:
            Union[np.ndarray, None]: X matrix
        """
        return self._X

    @property
    def y(self) -> Union[np.ndarray, None]:
        """property y

        Returns:
            Union[np.ndarray, None]: y matrix
        """
        return self._y

    @property
    def theta(self) -> Union[np.ndarray, None]:
        """property theta

        Returns:
            Union[np.ndarray, None]: theta matrix
        """
        return self._theta

    @property
    def theta_history(self) -> Union[list, None]:
        """theta training history

        Returns:
            Union[list, None]: theta history list
        """
        return self._theta_history

    @property
    def cost_history(self) -> Union[list, None]:
        """cost learning history

        Returns:
            Union[list, None]: cost history list
        """
        return self._cost_history

    def predict(self, X: np.ndarray) -> np.ndarray:
        """generate prediction

        Args:
            X (np.ndarray): input feature matrix

        Raises:
            ValueError: if shape is not proper for the input feature
            Warning: if model is not trained yet

        Returns:
            np.ndarray: predicted values
        """

        if self._theta is not None:
            format_x = self._format_x_for_theta_0(X)
            if format_x.shape[1] == self._theta.shape[0]:
                y_pred = format_x @ self._theta  # (m,1) = (m,n) * (n,1)
                return y_pred
            elif format_x.shape[1] == self._theta.shape[1]:
                y_pred = format_x @ self._theta.T  # (m,1) = (m,n) * (n,1)
                return y_pred
            else:
                raise ValueError("Shape is not proper.")
        else:
            raise Warning("Model is not trained yet. Theta is None.")

    def train(self,
              X: np.ndarray,
              y: np.ndarray,
              verbose: bool = True,
              method: str = "SGD",
              theta_precision: float = 0.001,
              batch_size: int = 30) -> None:
        """train model /theta estimator

        Args:
            X (np.ndarray): X matrix/feature matrix.
            y (np.ndarray): y matrix/target matrix.
            verbose (bool, optional): print things. Defaults to True.
            method (str, optional): training method. Defaults to "SGD".

                    Available-
                        "BGD"(Batch Graident Descent),
                        "SGD"(Stochastic Gradient Descent),
                        "NORMAL"(Normal Equation)

            theta_precision (float, optional): theta initialization value precision. Defaults to 0.001.
            batch_size (int, optional): batch size only for BGD. Defaults to 30.
        """

        self._X = self._format_x_for_theta_0(X)
        self._y = y

        # number of features+1 because of theta_0
        self._n = self._X.shape[1]
        self._m = self._y.shape[0]

        self._theta_history = []
        self._cost_history = []

        if method == "BGD":
            self._theta = np.random.rand(1, self._n) * theta_precision
            if verbose:
                print("random initial θ value :", self._theta)

            for iteration in range(self.iterations):
                # calculate y_pred
                y_pred = self.predict(self._X)

                # new θ to replace old θ
                new_theta = None

                # simultaneous operation
                gradient = np.mean((y_pred - self._y) * self._X, axis=0)
                new_theta = self._theta - (self.alpha * gradient)

                if np.isnan(np.sum(new_theta)) or np.isinf(np.sum(new_theta)):
                    print("breaking. found inf or nan.")
                    break
                # override with new θ
                self._theta = new_theta

                # calculate cost to put in history
                cost = calculate_mse_cost(y_pred=self.predict(X=self._X), y=self._y)
                self._cost_history.append(cost)

                # calcualted theta in history
                self._theta_history.append(self._theta[0])

        elif method == "SGD":  # stochastic gradient descent
            self._theta = np.random.rand(1, self._n) * theta_precision
            if verbose:
                print("random initial θ value :", self._theta)

            for iteration in range(self.iterations):

                # creating indices for batches
                indices = np.random.randint(0, self._m, size=batch_size)

                # creating batch for this iteration
                # X_batch = np.take(self._X, indices, axis=0)
                # y_batch = np.take(self._y, indices, axis=0)
                X_batch = self._X[indices,:]
                y_batch = self._y[indices,:]

                # calculate y_pred
                y_pred = self.predict(X_batch)
                # new θ to replace old θ
                new_theta = None

                # simultaneous operation
                gradient = np.mean((y_pred - y_batch) * X_batch, axis=0)
                new_theta = self._theta - (self.alpha * gradient)

                if np.isnan(np.sum(new_theta)) or np.isinf(np.sum(new_theta)):
                    print("breaking. found inf or nan.")
                    break
                # override with new θ
                self._theta = new_theta

                # calculate cost to put in history
                cost = calculate_mse_cost(y_pred=self.predict(X=X_batch), y=y_batch)
                self._cost_history.append(cost)

                # calcualted theta in history
                self._theta_history.append(self._theta[0])

        elif method == "NORMAL":
            self._theta = np.linalg.inv(self._X.T @ self._X) @ self._X.T @ self._y

        else:
            print("No Method Defined.")


class RidgeRegression:
    """Ridge Regression Model Class (L2 Regularization)

    Args:
        alpha (float, optional): learning rate. Defaults to 0.01.
        iterations (int, optional): number of iteratons. Defaults to 10000.

    References:
        https://machinelearningexploration.readthedocs.io/en/latest/LinearRegression/Explore.html#Ridge(L2-Regularization)-Regression
    """

    def __init__(self, alpha: float = 0.01, iterations: int = 10000):
        """Constructor
        """
        self.alpha = alpha
        self.iterations = iterations
        self._theta = None
        self._X = None
        self._y = None
        self._theta_history = None
        self._cost_history = None
        self._n = None
        self._m = None

    def _format_x_for_theta_0(self, x_i: np.ndarray) -> np.ndarray:
        """format X matrix for linear model

        put 1's in the first column as feature for theta_0

        Args:
            x_i (np.ndarray): input x matrix

        Returns:
            np.ndarray: formatted x matrix
        """
        x_i = x_i.copy()
        if len(x_i.shape) == 1:
            x_i = x_i.reshape(-1, 1)

        if False in (x_i[..., 0] == 1):
            return np.hstack(tup=(np.ones(shape=(x_i.shape[0], 1)), x_i))
        else:
            return x_i

    @property
    def X(self) -> Union[np.ndarray, None]:
        """property X

        Returns:
            Union[np.ndarray, None]: X matrix
        """
        return self._X

    @property
    def y(self) -> Union[np.ndarray, None]:
        """property y

        Returns:
            Union[np.ndarray, None]: y matrix
        """
        return self._y

    @property
    def theta(self) -> Union[np.ndarray, None]:
        """property theta

        Returns:
            Union[np.ndarray, None]: theta matrix
        """
        return self._theta

    @property
    def theta_history(self) -> Union[list, None]:
        """theta training history

        Returns:
            Union[list, None]: theta history list
        """
        return self._theta_history

    @property
    def cost_history(self) -> Union[list, None]:
        """cost learning history

        Returns:
            Union[list, None]: cost history list
        """
        return self._cost_history

    def predict(self, X: np.ndarray) -> np.ndarray:
        """generate prediction

        Args:
            X (np.ndarray): input feature matrix

        Raises:
            ValueError: if shape is not proper for the input feature
            Warning: if model is not trained yet

        Returns:
            np.ndarray: predicted values
        """

        if self._theta is not None:
            format_x = self._format_x_for_theta_0(X)
            if format_x.shape[1] == self._theta.shape[0]:
                y_pred = format_x @ self._theta  # (m,1) = (m,n) * (n,1)
                return y_pred
            elif format_x.shape[1] == self._theta.shape[1]:
                y_pred = format_x @ self._theta.T  # (m,1) = (m,n) * (n,1)
                return y_pred
            else:
                raise ValueError("Shape is not proper.")
        else:
            raise Warning("Model is not trained yet. Theta is None.")

    def train(self,
              X: np.ndarray,
              y: np.ndarray,
              verbose: bool = True,
              method: str = "SGD",
              theta_precision: float = 0.001,
              penalty: Union[float, int] = 1.0,
              batch_size: int = 30) -> None:
        """train model /theta estimator

        Args:
            X (np.ndarray): X matrix/feature matrix.
            y (np.ndarray): y matrix/target matrix.
            verbose (bool, optional): print things. Defaults to True.
            method (str, optional): training method. Defaults to "SGD".

                    Available-
                        "BGD"(Batch Graident Descent),
                        "SGD"(Stochastic Gradient Descent),
                        "NORMAL"(Normal Equation)

            theta_precision (float, optional): theta initialization value precision. Defaults to 0.001.
            penalty (Union[float, int], optional): regularization penalty. Defaults to 1.0.
            batch_size (int, optional): batch size only for BGD. Defaults to 30.
        """

        self._X = self._format_x_for_theta_0(X)
        self._y = y

        # number of features+1 because of theta_0
        self._n = self._X.shape[1]
        self._m = self._y.shape[0]

        self._theta_history = []
        self._cost_history = []

        if method == "BGD":
            self._theta = np.random.rand(1, self._n) * theta_precision
            if verbose:
                print("random initial θ value :", self._theta)

            for iteration in range(self.iterations):
                # calculate y_pred
                y_pred = self.predict(self._X)
                # new θ to replace old θ
                new_theta = None

                # simultaneous operation
                ######################################################

                gradient = np.mean((y_pred - self._y) * self._X, axis=0)

                # theta_0 will not be effected by penalty
                new_theta_0 = self._theta[:, [0]] - (self.alpha * gradient[0])
                # rest of theta's will be effected by it
                new_theta_rest = self._theta[:, range(
                    1, self._n)] * (1 - (penalty/self._m)) - (self.alpha * gradient[1:])

                new_theta = np.hstack((new_theta_0, new_theta_rest))

                if np.isnan(np.sum(new_theta)) or np.isinf(np.sum(new_theta)):
                    print("breaking. found inf or nan.")
                    break
                # override with new θ
                self._theta = new_theta

                # calculate cost to put in history
                cost = calculate_mse_cost(y_pred=self.predict(X=self._X), y=self._y)
                self._cost_history.append(cost)

                # calcualted theta in history
                self._theta_history.append(self._theta[0])
        elif method == "SGD":
            self._theta = np.random.rand(1, self._n) * theta_precision
            if verbose:
                print("random initial θ value :", self._theta)

            for iteration in range(self.iterations):

                indices = np.random.randint(0, self._m, size=batch_size)

                # X_batch = np.take(self._X, indices, axis=0)
                # y_batch = np.take(self._y, indices, axis=0)

                X_batch = self._X[indices,:]
                y_batch = self._y[indices,:]

                # calculate y_pred
                y_pred = self.predict(X_batch)
                # new θ to replace old θ
                new_theta = None

                # simultaneous operation
                gradient = np.mean((y_pred - y_batch) * X_batch, axis=0)
                new_theta_0 = self._theta[:,[0]] - (self.alpha * gradient[0])
                new_theta_rest = self._theta[:,range(1,self._n)] * (1 - (penalty/self._m) ) - (self.alpha * gradient[1:])

                new_theta = np.hstack((new_theta_0,new_theta_rest))


                if np.isnan(np.sum(new_theta)) or np.isinf(np.sum(new_theta)):
                    print("breaking. found inf or nan.")
                    break
                # override with new θ
                self._theta = new_theta

                # calculate cost to put in history
                cost = calculate_mse_cost(y_pred=self.predict(X=X_batch), y=y_batch)
                self._cost_history.append(cost)

                # calcualted theta in history
                self._theta_history.append(self._theta[0])

        elif method == "NORMAL":
            self._theta = np.linalg.inv(
                self._X.T @ self._X + (penalty * np.identity(self._n))) @ self._X.T @ self._y

        else:
            print("No Method Defined.")


class LassoRegression:
    """Lasso Regression Model Class (L1 Regularization)

    Args:
        alpha (float, optional): learning rate. Defaults to 0.01.
        iterations (int, optional): number of iteratons. Defaults to 10000.

    References:
        https://machinelearningexploration.readthedocs.io/en/latest/LinearRegression/Explore.html#Lasso(L1-Regularization)-Regression
    """

    def __init__(self, alpha: float = 0.01, iterations: int = 10000):
        """Constructor
        """
        self.alpha = alpha
        self.iterations = iterations
        self._theta = None
        self._X = None
        self._y = None
        self._theta_history = None
        self._cost_history = None
        self._n = None
        self._m = None

    def _format_x_for_theta_0(self, x_i: np.ndarray) -> np.ndarray:
        """format X matrix for linear model

        put 1's in the first column as feature for theta_0

        Args:
            x_i (np.ndarray): input x matrix

        Returns:
            np.ndarray: formatted x matrix
        """
        x_i = x_i.copy()
        if len(x_i.shape) == 1:
            x_i = x_i.reshape(-1, 1)

        if False in (x_i[..., 0] == 1):
            return np.hstack(tup=(np.ones(shape=(x_i.shape[0], 1)), x_i))
        else:
            return x_i

    @property
    def X(self) -> Union[np.ndarray, None]:
        """property X

        Returns:
            Union[np.ndarray, None]: X matrix
        """
        return self._X

    @property
    def y(self) -> Union[np.ndarray, None]:
        """property y

        Returns:
            Union[np.ndarray, None]: y matrix
        """
        return self._y

    @property
    def theta(self) -> Union[np.ndarray, None]:
        """property theta

        Returns:
            Union[np.ndarray, None]: theta matrix
        """
        return self._theta

    @property
    def theta_history(self) -> Union[list, None]:
        """theta training history

        Returns:
            Union[list, None]: theta history list
        """
        return self._theta_history

    @property
    def cost_history(self) -> Union[list, None]:
        """cost learning history

        Returns:
            Union[list, None]: cost history list
        """
        return self._cost_history

    def predict(self, X: np.ndarray) -> np.ndarray:
        """generate prediction

        Args:
            X (np.ndarray): input feature matrix

        Raises:
            ValueError: if shape is not proper for the input feature
            Warning: if model is not trained yet

        Returns:
            np.ndarray: predicted values
        """

        if self._theta is not None:
            format_x = self._format_x_for_theta_0(X)
            if format_x.shape[1] == self._theta.shape[0]:
                y_pred = format_x @ self._theta  # (m,1) = (m,n) * (n,1)
                return y_pred
            elif format_x.shape[1] == self._theta.shape[1]:
                y_pred = format_x @ self._theta.T  # (m,1) = (m,n) * (n,1)
                return y_pred
            else:
                raise ValueError("Shape is not proper.")
        else:
            raise Warning("Model is not trained yet. Theta is None.")

    def train(self,
              X: np.ndarray,
              y: np.ndarray,
              verbose: bool = True,
              method: str = "SGD",
              theta_precision: float = 0.001,
              penalty: Union[int, float] = 1.0,
              batch_size: int = 30) -> None:
        """train model /theta estimator

        Args:
            X (np.ndarray): X matrix/feature matrix.
            y (np.ndarray): y matrix/target matrix.
            verbose (bool, optional): print things. Defaults to True.
            method (str, optional): training method. Defaults to "SGD".

                    Available-
                        "BGD"(Batch Graident Descent),
                        "SGD"(Stochastic Gradient Descent)

            theta_precision (float, optional): theta initialization value precision. Defaults to 0.001.
            penalty (Union[float, int], optional): regularization penalty. Defaults to 1.0.
            batch_size (int, optional): batch size only for BGD. Defaults to 30.
        """

        self._X = self._format_x_for_theta_0(X)
        self._y = y

        # number of features+1 because of theta_0
        self._n = self._X.shape[1]
        self._m = self._y.shape[0]

        self._theta_history = []
        self._cost_history = []

        if method == "BGD":
            self._theta = np.random.rand(1, self._n) * theta_precision
            if verbose:
                print("random initial θ value :", self._theta)

            for iteration in range(self.iterations):
                # calculate y_pred
                y_pred = self.predict(self._X)
                # new θ to replace old θ
                new_theta = np.zeros_like(self._theta)

                # simultaneous operation
                ################################################################################################################

                gradient = np.mean((y_pred - self._y) * self._X, axis=0)
                new_theta_0 = self._theta[:, [0]] - (self.alpha * gradient[0])
                new_theta_rest = self._theta[:, range(
                    1, self._n)] - (self.alpha * gradient[1:]) - (penalty/self._m)

                new_theta = np.hstack((new_theta_0, new_theta_rest))

                if np.isnan(np.sum(new_theta)) or np.isinf(np.sum(new_theta)):
                    print("breaking. found inf or nan.")
                    break
                # override with new θ
                self._theta = new_theta

                # calculate cost to put in history
                cost = calculate_mse_cost(y_pred=self.predict(X=self._X), y=self._y)
                self._cost_history.append(cost)

                # calcualted theta in history
                self._theta_history.append(self._theta[0])

        elif method == "SGD":
            self._theta = np.random.rand(1, self._n) * theta_precision
            if verbose:
                print("random initial θ value :", self._theta)

            for iteration in range(self.iterations):

                indices = np.random.randint(0, self._m, size=batch_size)

                # X_batch = np.take(self._X, indices, axis=0)
                # y_batch = np.take(self._y, indices, axis=0)
                X_batch = self._X[indices, :]
                y_batch = self._y[indices, :]

                # calculate y_pred
                y_pred = self.predict(X_batch)
                # new θ to replace old θ
                new_theta = None

                # simultaneous operation
                ################################################################################################################

                gradient = np.mean((y_pred - y_batch) * X_batch, axis=0)
                new_theta_0 = self._theta[:, [0]] - (self.alpha * gradient[0])
                new_theta_rest = self._theta[:, range(
                    1, self._n)] - (self.alpha * gradient[1:]) - (penalty/self._m)

                new_theta = np.hstack((new_theta_0, new_theta_rest))

                if np.isnan(np.sum(new_theta)) or np.isinf(np.sum(new_theta)):
                    print("breaking. found inf or nan.")
                    break
                # override with new θ
                self._theta = new_theta

                # calculate cost to put in history
                cost = calculate_mse_cost(y_pred=self.predict(X=X_batch), y=y_batch)
                self._cost_history.append(cost)

                # calcualted theta in history
                self._theta_history.append(self._theta[0])

        else:
            print("No Method Defined.")


class LogisticRegression:
    """Logisitic Regression Model Class

    Args:
        alpha (float, optional): [description]. Defaults to 0.01.
        iterations (int, optional): [description]. Defaults to 10000.

    References:
        https://machinelearningexploration.readthedocs.io/en/latest/LogisticRegression/Explore.html
    """

    def __init__(self, alpha: float = 0.01, iterations: int = 10000):
        self.alpha = alpha
        self.iterations = iterations
        self._theta = None
        self._X = None
        self._y = None
        self._theta_history = None
        self._cost_history = None
        self._n = None
        self._m = None

    def _format_x_for_theta_0(self, x_i: np.ndarray) -> np.ndarray:
        """format X matrix for linear model

        put 1's in the first column as feature for theta_0

        Args:
            x_i (np.ndarray): input x matrix

        Returns:
            np.ndarray: formatted x matrix
        """

        x_i = x_i.copy()
        if len(x_i.shape) == 1:
            x_i = x_i.reshape(-1, 1)

        if False in (x_i[..., 0] == 1):
            return np.hstack(tup=(np.ones(shape=(x_i.shape[0], 1)), x_i))
        else:
            return x_i

    @property
    def X(self) -> Union[np.ndarray, None]:
        """property X

        Returns:
            Union[np.ndarray, None]: X matrix
        """
        return self._X

    @property
    def y(self) -> Union[np.ndarray, None]:
        """property y

        Returns:
            Union[np.ndarray, None]: y matrix
        """
        return self._y

    @property
    def theta(self) -> Union[np.ndarray, None]:
        """property theta

        Returns:
            Union[np.ndarray, None]: theta matrix
        """
        return self._theta

    @property
    def theta_history(self) -> Union[list, None]:
        """theta training history

        Returns:
            Union[list, None]: theta history list
        """
        return self._theta_history

    @property
    def cost_history(self) -> Union[list, None]:
        """cost learning history

        Returns:
            Union[list, None]: cost history list
        """
        return self._cost_history

    def predict(self, X: np.ndarray) -> np.ndarray:
        """generate prediction

        Args:
            X (np.ndarray): input feature matrix

        Raises:
            ValueError: if shape is not proper for the input feature
            Warning: if model is not trained yet

        Returns:
            np.ndarray: predicted values
        """
        if self._theta is not None:
            format_X = self._format_x_for_theta_0(X)

            if format_X.shape[1] == self._theta.shape[0]:
                y_pred = sigmoid(format_X @ self._theta)  # (m,1) = (m,n) * (n,1)
                return y_pred
            elif format_X.shape[1] == self._theta.shape[1]:
                y_pred = sigmoid(format_X @ self._theta.T)  # (m,1) = (m,n) * (n,1)
                return y_pred
            else:
                raise ValueError("Shape is not proper.")
        else:
            raise Warning("Model is not trained yet. Theta is None.")

    def train(self,
              X: np.ndarray,
              y: np.ndarray,
              verbose: bool = True,
              method: str = "SGD",
              theta_precision: float = 0.001,
              batch_size: int = 30,
              regularization: bool = False,
              penalty: Union[float, int] = 1.0) -> None:
        """train theta / estimator

        Args:
            X (np.ndarray): X matrix/feature matrix.
            y (np.ndarray): y matrix/target matrix.
            verbose (bool, optional): print things. Defaults to True.
            method (str, optional): training method. Defaults to "SGD".

                    Available-
                        "BGD"(Batch Graident Descent),
                        "SGD"(Stochastic Gradient Descent)

            theta_precision (float, optional): theta initialization value precision. Defaults to 0.001.
            batch_size (int, optional): batch size only for BGD. Defaults to 30.            
            regularization (bool, optional): Apply Regularization. Defaults to False.
            penalty (Union[float, int], optional): regularization penalty only works for regularization=True. Defaults to 1.0.
        """

        self._X = self._format_x_for_theta_0(X)
        self._y = y

        # number of features+1 because of theta_0
        self._n = self._X.shape[1]
        self._m = self._y.shape[0]

        self._theta_history = []
        self._cost_history = []

        if method == "BGD":
            self._theta = np.random.rand(1, self._n) * theta_precision
            if verbose:
                print("random initial θ value :", self._theta)

            for iteration in range(self.iterations):
                # calculate y_pred
                y_pred = self.predict(self._X)

                # new θ to replace old θ
                new_theta = None

                # simultaneous operation

                if regularization:
                    gradient = np.mean((y_pred - self._y) * self._X, axis=0)
                    new_theta_0 = self._theta[:, [0]] - (self.alpha * gradient[0])
                    new_theta_rest = self._theta[:, range(
                        1, self._n)] * (1 - (penalty/self._m)) - (self.alpha * gradient[1:])
                    new_theta = np.hstack((new_theta_0, new_theta_rest))

                else:
                    gradient = np.mean((y_pred - self._y) * self._X, axis=0)
                    new_theta = self._theta - (self.alpha * gradient)

                if np.isnan(np.sum(new_theta)) or np.isinf(np.sum(new_theta)):
                    print("breaking. found inf or nan.")
                    break
                # override with new θ
                self._theta = new_theta

                # calculate cost to put in history
                cost = calculate_entropy_cost(y_pred=self.predict(X=self._X), y=self._y)
                self._cost_history.append(cost)

                # calcualted theta in history
                self._theta_history.append(self._theta[0])

        elif method == "SGD":  # stochastic gradient descent
            self._theta = np.random.rand(1, self._n) * theta_precision
            if verbose:
                print("random initial θ value :", self._theta)

            for iteration in range(self.iterations):

                # creating indices for batches
                indices = np.random.randint(0, self._m, size=batch_size)

                # creating batch for this iteration
                # X_batch = np.take(self._X, indices, axis=0)
                # y_batch = np.take(self._y, indices, axis=0)
                X_batch = self._X[indices, :]
                y_batch = self._y[indices, :]

                # calculate y_pred
                y_pred = self.predict(X_batch)
                # new θ to replace old θ
                new_theta = None

                # simultaneous operation
                if regularization:
                    gradient = np.mean((y_pred - y_batch) * X_batch, axis=0)
                    new_theta_0 = self._theta[:, [0]] - (self.alpha * gradient[0])
                    new_theta_rest = self._theta[:, range(
                        1, self._n)] * (1 - (penalty/self._m)) - (self.alpha * gradient[1:])
                    new_theta = np.hstack((new_theta_0, new_theta_rest))

                else:
                    gradient = np.mean((y_pred - y_batch) * X_batch, axis=0)
                    new_theta = self._theta - (self.alpha * gradient)

                if np.isnan(np.sum(new_theta)) or np.isinf(np.sum(new_theta)):
                    print("breaking. found inf or nan.")
                    break
                # override with new θ
                self._theta = new_theta

                # calculate cost to put in history
                cost = calculate_entropy_cost(y_pred=self.predict(X=X_batch), y=y_batch)
                self._cost_history.append(cost)

                # calcualted theta in history
                self._theta_history.append(self._theta[0])

        else:
            print("No Method Defined.")


def polynomial_regression(x: np.ndarray, y: np.ndarray, degree: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    fit Regression line with polynomial degree.

    Args:
        x (np.ndarray): independent variable.
        y (np.ndarray): dependent variable.
        degree (int): polynomial degree.

    Returns:
        Tuple[np.ndarray, np.ndarray, np.ndarray]: slope, residual, fitline.

    Examples:
        >>> import matplotlib.pyplot as plt
        >>> x = np.arange(-10, 10)
        >>> y = x**2 + x**3
        >>> s, r, l = polynomial_regression(x, y, 3)
        >>> plt.plot(x, y, 'ko', label='original')
        >>> plt.plot(x, l, '.-',  label='regression line')
        >>> plt.legend()
        >>> plt.show()
    """
    a = np.polyfit(x, y, degree)
    slope = a[:-1]
    resid = a[-1:]
    fit_line = np.array([(x**(degree - i))*slope[i]
                     for i in range(0, degree)]).sum(axis=0) + resid
    return slope, resid, fit_line


def trend(x: np.ndarray, y: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    get trend of the data.

    Args:
        x (np.ndarray): independent variable.
        y (np.ndarray): dependent variable.

    Returns:
        Tuple[np.ndarray, np.ndarray, np.ndarray]: slope, residual, trendline. 

    Examples;
        >>> import matplotlib.pyplot as plt 
        >>> x = np.array([1, 2, 3, 4, 5, 6, 7, 8])
        >>> y = np.array([1, 2, 3, 3, 4, 5, 7, 10])
        >>> s, r, t = trend(x, y)
        >>> plt.plot(x, y, 'o', label='original')
        >>> plt.plot(x, t, '.-',  label='regression line')
        >>> plt.legend()
        >>> plt.show()
    """
    return polynomial_regression(x, y, 1)


if __name__ == "__main__":
    # import matplotlib.pyplot as plt

    # x = np.array([1, 2, 3, 4, 5, 6, 7, 8])
    # y = np.array([1, 2, 3, 3, 4, 5, 7, 10])
    # s, r, t = trend(x, y)
    # plt.plot(x, y, 'o', label='original')
    # plt.plot(x, t, '.-',  label='regression line')
    # plt.legend()
    # plt.show()


    # x = np.arange(-10, 10)
    # y = x**2 + x**3
    # s, r, l = polynomial_regression(x, y, 3)

    # plt.plot(x, y, 'ko', label='original')
    # plt.plot(x, l, '.-',  label='regression line')
    # plt.legend()
    # plt.show()

    pass



"""
Ensemble methods for Machine Learning
"""
from __future__ import annotations
from typing import Union, Tuple, List, Optional
import numpy as np
from mightypy.ml._tree import DecisionTreeClassifier, DecisionTreeRegressor


class RandomForestClassifier:
    """Ensemble method for classification 

    using a bunch of Decision Tree's to do to the classification.

    Args:
        num_of_trees (int, optional): number of trees in ensemble. Defaults to 50.
        min_features (int, optional): minimum number of features to use in every tree. Defaults to None.
        max_depth (int, optional): max depth of the every tree. Defaults to 100.
        min_samples_split (int, optional): minimum size ofsampels to split. Defaults to 2.
        criteria (str, optional): criteria to calcualte information gain. Defaults to 'gini'.
    """

    def __init__(self, num_of_trees: int = 25, min_features: Optional[int] = None, max_depth: int = 50, min_samples_split: int = 2, criteria: str = 'gini') -> None:
        """constructor
        """
        self._X = None
        self._y = None
        self._feature_names = None
        self._target_name = None
        self._trees = []
        self.num_of_trees = num_of_trees
        self.min_features = min_features
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.criteria = criteria

    def _sampling(self) -> Tuple[List[int], List[int]]:
        """sampling function

        Returns:
            Tuple[List[int],List[int]]: sampling idxs for rows nad columns for feature and target matrix.
        """
        m, n = self._X.shape

        # sampling with replacement
        # means rows with repeat in the data
        # statitistically it gives an edge for data prediction
        idxs = np.random.randint(low=0, high=m, size=m)

        # feature sampling to decrease correlation between trees
        if self.min_features is None:
            size = n
        else:
            size = n if self.min_features > n else self.min_features

        feat_idxs = np.random.choice(n, size=size, replace=False)
        return idxs, feat_idxs

    def train(self, X: Union[np.ndarray, list], y: Union[np.ndarray, list], feature_name: list = None, target_name: list = None) -> None:
        """Train the model

        Args:
            X (Union[np.ndarray,list]): feature matrix
            y (Union[np.ndarray,list]): target matrix
            feature_name (str, optional): feature names. Defaults to None.
            target_name (str, optional): target names. Defaults to None.
        """

        X = np.array(X, dtype='O') if not isinstance(
            X, (np.ndarray)) else X  # converting to numpy array
        y = np.array(y, dtype='O') if not isinstance(
            y, (np.ndarray)) else y  # converting to numpy array
        # reshaping to vectors
        self._X = X.reshape(-1, 1) if len(X.shape) == 1 else X
        self._y = y.reshape(-1, 1) if len(y.shape) == 1 else y

        # creating feature names if not mentioned
        self._feature_names = feature_name or [
            f"C_{i}" for i in range(self._X.shape[1])]

        # creating target name if not mentioned
        self._target_name = target_name or ['target']

        for _ in range(self.num_of_trees):
            clf = DecisionTreeClassifier(
                max_depth=self.max_depth,
                min_samples_split=self.min_samples_split,
                criteria=self.criteria
            )
            idxs, feat_idxs = self._sampling()
            X_sampled = self._X[idxs, :][:, feat_idxs]
            y_sampled = self._y[idxs]

            clf.train(
                X=X_sampled,
                y=y_sampled,
                feature_name=[self._feature_names[i] for i in feat_idxs],
                target_name=self._target_name
            )

            self._trees.append([clf, feat_idxs])

    @staticmethod
    def _get_max_result(a: np.ndarray) -> Union[str, int, None]:
        """get max result from the bunch of classification results

        Args:
            a (np.ndarray): input array for category

        Returns:
            Union[str,int,None]: max count class/category
        """
        unique_values = np.unique(a, return_counts=True)
        zipped = zip(*unique_values)
        max_count = 0
        result = None
        for i in zipped:
            if i[1] > max_count:
                result = i[0]
                max_count = i[1]
        return result

    def predict(self, X: np.ndarray) -> np.ndarray:
        """predict results

        Args:
            X (np.ndarray): test matrix.

        Raises:
            ValueError: X should be list or numpy array.

        Returns:
            np.ndarray: prediction results.
        """

        if isinstance(X, (np.ndarray, list)):
            X = np.array(X, dtype='O') if not isinstance(X, (np.ndarray)) else X

            results = []
            for clf, feat_idxs in self._trees:
                result = clf.predict(X[:, feat_idxs])
                results.append(result)

            all_tree_results = np.concatenate(np.array(results, dtype='O'), axis=1)
            final_results = np.apply_along_axis(
                func1d=self._get_max_result, axis=1, arr=all_tree_results).reshape(-1, 1)
            return final_results
        else:
            raise ValueError("X should be list or numpy array")


class RandomForestRegressor:
    """Ensemble method for regression 

    using a bunch of Decision Tree's to do to the regression.

    Args:
        num_of_trees (int, optional): number of trees in ensemble. Defaults to 50.
        min_features (int, optional): minimum number of features to use in every tree. Defaults to None.
        max_depth (int, optional): max depth of the every tree. Defaults to 100.
        min_samples_split (int, optional): minimum size ofsampels to split. Defaults to 2.
        criteria (str, optional): criteria to calcualte information gain. Defaults to 'gini'.
    """

    def __init__(self, num_of_trees: int = 25, min_features: Opional[int] = None, max_depth: int = 30, min_samples_split: int = 3, criteria: str = 'variance') -> None:
        self._X = None
        self._y = None
        self._feature_names = None
        self._target_name = None
        self._trees = []
        self.num_of_trees = num_of_trees
        self.min_features = min_features
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.criteria = criteria

    def _sampling(self) -> Tuple[List[int], List[int]]:
        """sampling function

        Returns:
            Tuple[List[int],List[int]]: sampling idxs for rows nad columns for feature and target matrix.
        """
        m, n = self._X.shape

        # sampling with replacement
        # means rows with repeat in the data
        # statitistically it gives an edge for data prediction
        idxs = np.random.randint(low=0, high=m, size=m)

        # feature sampling to decrease correlation between trees
        if self.min_features is None:
            size = n
        else:
            size = n if self.min_features > n else self.min_features

        feat_idxs = np.random.choice(n, size=size, replace=False)
        return idxs, feat_idxs

    def train(self, X: Union[np.ndarray, list], y: Union[np.ndarray, list], feature_name: list = None, target_name: list = None) -> None:
        """Train the model

        Args:
            X (Union[np.ndarray,list]): feature matrix.
            y (Union[np.ndarray,list]): target matrix.
            feature_name (list, optional): feature names. Defaults to None.
            target_name (list, optional): target name. Defaults to None.
        """

        X = np.array(X, dtype='O') if not isinstance(
            X, (np.ndarray)) else X  # converting to numpy array
        y = np.array(y, dtype='O') if not isinstance(
            y, (np.ndarray)) else y  # converting to numpy array
        # reshaping to vectors
        self._X = X.reshape(-1, 1) if len(X.shape) == 1 else X
        self._y = y.reshape(-1, 1) if len(y.shape) == 1 else y

        # creating feature names if not mentioned
        self._feature_names = feature_name or [
            f"C_{i}" for i in range(self._X.shape[1])]

        # creating target name if not mentioned
        self._target_name = target_name or ['target']

        for _ in range(self.num_of_trees):

            reg = DecisionTreeRegressor(
                max_depth=self.max_depth,
                min_samples_split=self.min_samples_split,
                criteria=self.criteria
            )
            idxs, feat_idxs = self._sampling()  # get sampling idxs
            X_sampled = self._X[idxs, :][:, feat_idxs]
            y_sampled = self._y[idxs]

            reg.train(
                X=X_sampled,
                y=y_sampled,
                feature_name=[self._feature_names[i] for i in feat_idxs],
                target_name=self._target_name
            )

            self._trees.append([reg, feat_idxs])

    def predict(self, X: Union[np.ndarray, list]) -> np.ndarray:
        """predict regression result

        Args:
            X (Union[np.ndarray, list]): test matrix.

        Raises:
            ValueError: X should be list or numpy array.

        Returns:
            np.ndarray: regression results.
        """
        if isinstance(X, (np.ndarray, list)):
            X = np.array(X, dtype='O') if not isinstance(X, (np.ndarray)) else X

            results = []
            for reg, feat_idxs in self._trees:
                result = reg.predict(X[:, feat_idxs])
                results.append(result)

            all_tree_results = np.concatenate(np.array(results, dtype='O'), axis=1)
            final_results = np.mean(all_tree_results, axis=1).reshape(-1, 1)
            return final_results
        else:
            raise ValueError("X should be list or numpy array")


class AdaboostClassifier:
    """
    Adaboost Classification Model.
    
    It is still under construction. DO NOT USE IT.

    .. _Adaboost: https://machinelearningexploration.readthedocs.io/en/latest/EnsembleMethods/ExploreBoosting.html#Adaboost-Classfication
    """
    def __init__(self, n_stumps:int, stump_depth:int=0):
        print("DO not use this method. It is not tested.")
        self.stump_depth = stump_depth
        self._X= None
        self._y= None
        self._feature_names = None
        self._stumps:list = []
        self.n_stumps = n_stumps

    def amount_of_say(self, total_err:Union[np.ndarray, float, int])->Union[np.ndarray, float, int]:
        """
        Amount of say.

        .. math::
            \\text{amount of say} = \\frac{1}{2}(log(\\frac{1 - \\text{TE}}{\\text{TE}}))

        Args:
            total_err (Union[np.ndarray, float, int]): Total error from tree.

        Returns:
            Union[np.ndarray, float, int]: amount of say.
        """
        return np.log((1 - total_err)/ total_err) / 2.0

    def normalize(self, x:np.ndarray)->np.ndarray:
        """
        Nornmalization.

        Args:
            x (np.ndarray): input array.

        Returns:
            np.ndarray: normalized values.
        """
        return x / x.sum()

    def _update_sample_weight(self, sample_weights:Union[np.ndarray,float], aos:Union[int,float], is_correct:bool)->np.ndarray:
        """
        Update sample weight for new tree.

        Args:
            sample_weights (Union[np.ndarray,float]): sample weight.
            aos (Union[int,float]): amount of say.
            is_correct (bool): is correctly classified records.

        Returns:
            Union[np.ndarray,float]: New updated weights based on amount of say.
        """
        if is_correct: # correctly classified records
            return sample_weights * np.exp(aos)
        else:
            return sample_weights * np.exp(-aos)

    def _get_sample_idxs(self, size:int, probs:Union[list,np.ndarray])->np.ndarray:
        """
        get indexes based on probabilities from a range.

        Args:
            size (int): size of array.
            probs (Union[list,np.ndarray]): probabities of indexes.

        Returns:
            np.ndarray: sampling indexes.
        """
        idxs = np.random.choice(range(size),size=(size,), p=probs)
        return idxs 

    def predict(self, X: np.ndarray) -> np.ndarray:
        """
        Generate predictions.

        Args:
            X (np.ndarray): input array.

        Returns:
            np.ndarray: predictions.
        """
        results = np.ones((X.shape[0], 1)) * self.leaf_value
        for stump in self._stumps:
            results = results + (stump['aos'] * stump['model'].predict(X))

        return results

    def train(self, X:np.ndarray, y:np.ndarray, feature_names:list=None, target_name:list=None):
        """
        Train the model.

        Args:
            X (np.ndarray): input features.
            y (np.ndarray): target.
            feature_names (list, optional): feature names. Defaults to None.
            target_name (list, optional): target name. Defaults to None.
        """
        X = np.array(X, dtype='O') if not isinstance(
            X, (np.ndarray)) else X  # converting to numpy array
        y = np.array(y, dtype='O') if not isinstance(
            y, (np.ndarray)) else y  # converting to numpy array
        # reshaping to vectors
        self._X = X.reshape(-1, 1) if len(X.shape) == 1 else X
        self._y = y.reshape(-1, 1) if len(y.shape) == 1 else y

        # creating feature names if not mentioned
        self._feature_names = feature_names or [
            f"C_{i}" for i in range(self._X.shape[1])]

        # creating target name if not mentioned
        self._target_name = target_name or ['target']

        self.leaf_value = 1 / X.shape[0]
        sample_weights = np.ones((X.shape[0], 1)) * self.leaf_value

        sample_X = self._X
        sample_y = self._y
        self._n_samples = sample_X.shape[0]
        aos = -1

        for i_stump in range(self.n_stumps):
            print(i_stump," number of stump")
            
            # building a decision tree stump
            stump = DecisionTreeClassifier(max_depth=self.stump_depth)
            stump.train(
                X=sample_X,
                y=sample_y,
                feature_name=self._feature_names,
                target_name=self._target_name
            )

            stump_preds = self.predict(sample_X) + stump.predict(sample_X)
            
            total_err = ((stump_preds != sample_y) * sample_weights).sum()

            if total_err <= 0:
                print("early stopping as total error is <= 0", total_err)
                break

            aos = self.amount_of_say(total_err)
            
            if aos <= 0.0:
                print("early stopping as amount of say is <= 0", aos)
                break

            # storing in bag of stumps
            self._stumps.append({
                "idx": i_stump,
                "model": stump,
                "aos": aos
            })

            # preparation for next stump
            wrong_class_weights = (stump_preds != sample_y) * sample_weights
            right_class_weights = (stump_preds == sample_y) * sample_weights


            new_wrong_class_weights = self._update_sample_weight(
                sample_weights=wrong_class_weights,
                aos = aos,
                is_correct= False
            )

            new_right_class_weights = self._update_sample_weight(
                sample_weights=right_class_weights,
                aos = aos,
                is_correct=True
            )
            
            # new sample weights
            new_sample_weights = new_right_class_weights + new_wrong_class_weights
            sample_weights = self.normalize(new_sample_weights)

            # new samples
            sample_idxs = self._get_sample_idxs(size=self._n_samples,probs=sample_weights[...,-1])
            sample_X = sample_X[sample_idxs]
            sample_y = sample_y[sample_idxs]


    
    

if __name__ == "__main__":
    from sklearn.datasets import make_classification

    X,y = make_classification(n_classes=2,n_features=4,n_samples=100)

    model = AdaboostClassifier(n_stumps=100)
    model.train(
        X=X,
        y=y
    )
    print(model.predict(X))



"""
Signal Processing module.


author : Nishant Baheti <nishantbaheti.it19@gmail.com>
"""

import numpy as np

class FFTMixins:
    @staticmethod
    def psd(f_hat, l_signal):
        """Power Spectral Density."""
        return ((f_hat * np.conjugate(f_hat)) / l_signal).real  # type: ignore

    @staticmethod
    def magnitude(f_hat, l_signal):
        """Magnitude."""
        return (np.abs(f_hat) / l_signal).real  # type: ignore


class DenoiseFFT(FFTMixins):
    """
    Denoise signals with Fast Fourier method.

    Args:
        method (str): method. accepted values are psd, mag.
        threshold (float): threshold for cleanup.

    References:
        https://machinelearningexploration.readthedocs.io/en/latest/MathExploration/SignalProcessingFFT.html

    Examples:
        >>> import matplotlib.pyplot as plt
        >>> from mightypy.make import sine_wave_from_timesteps
        >>> time_step = 0.001
        >>> wave1, time1, freqs1 = sine_wave_from_timesteps(signal_freq=50, time_step=time_step)
        >>> wave2, time2, freqs2 = sine_wave_from_timesteps(signal_freq=70, time_step=time_step)
        >>> original_signal = wave1 + wave2
        >>> N = len(original_signal)
        >>> noisy_signal = original_signal + 2.5 * np.random.randn(N) + 2.8 * np.random.randn(N)  # adding random noise here
        >>> model = DenoiseFFT('psd', 100)
        >>> cleaned_signal = model.transform(noisy_signal)
        >>> plt.plot(original_signal, label='original')
        >>> plt.plot(noisy_signal, label='noisy')
        >>> plt.plot(cleaned_signal, label='cleaned')
        >>> plt.legend(loc='best')
        >>> plt.show()
    """

    def __init__(self, method: str, threshold: float) -> None:
        assert method.lower() in ('psd', 'mag'), "denoise signal method should be in psd, mag."
        self._method = method.lower()
        self._threshold = threshold
        super().__init__()

    def transform(self, signal: np.ndarray) -> np.ndarray:
        """
        Perform denoising operation on signal.

        Args:
            signal (np.ndarray): signal.

        Returns:
            np.ndarray: cleaned signal.
        """
        self.signal = signal
        self._l_signal = len(self.signal)
        self._f_hat = np.fft.fft(self.signal, self._l_signal)
        
        if self._method == 'psd':
            x = super().psd(self._f_hat, self._l_signal)
        else:
            x = super().magnitude(self._f_hat, self._l_signal)

        above_thresh_flag = x > self._threshold
        cleaned_f_hat = self._f_hat * above_thresh_flag

        return np.fft.ifft(cleaned_f_hat).real  # type: ignore


if __name__ == '__main__':

    import matplotlib.pyplot as plt
    from mightypy.make import sine_wave_from_timesteps

    time_step = 0.001
    wave1, time1, freqs1 = sine_wave_from_timesteps(signal_freq=50, time_step=time_step)
    wave2, time2, freqs2 = sine_wave_from_timesteps(signal_freq=70, time_step=time_step)
    original_signal = wave1 + wave2

    N = len(original_signal)

    noisy_signal = original_signal + 2.5 * np.random.randn(N) + 2.8 * np.random.randn(N)  # adding random noise here

    model = DenoiseFFT('psd', 100)
    cleaned_signal = model.transform(noisy_signal)

    plt.plot(original_signal, label='original')
    plt.plot(noisy_signal, label='noisy')
    plt.plot(cleaned_signal, label='cleaned')
    plt.legend(loc='best')
    plt.show()

    model = DenoiseFFT('mag', 0.2)
    cleaned_signal = model.transform(noisy_signal)

    plt.plot(original_signal, label='original')
    plt.plot(noisy_signal, label='noisy')
    plt.plot(cleaned_signal, label='cleaned')
    plt.legend(loc='best')
    plt.show()



"""
mightypy.stats
================

"""

from ._feature_importance import woe_and_iv

from ._data_drift import population_stability_index


__all__ = [
    "woe_and_iv",
    "population_stability_index"
]


"""
Dataset generation module.

author : Nishant Baheti <nishantbaheti.it19@gmail.com>
"""

import numpy as np
from typing import Tuple


def rotation_matrix_2d(theta: float) -> np.ndarray:
    """
    Create 2D data rotation matrix.

    Reference article
    ------------------
    
    https://en.wikipedia.org/wiki/Rotation_matrix

    Args:
        theta (float): angle for rotation.

    Returns:
        np.ndarray: rotation matrix.
    """

    mat = np.array([
        [np.cos(theta), -np.sin(theta)],
        [np.sin(theta), np.cos(theta)]
    ])
    return mat


def spiral_data(data_limit: int = 30, n_classes: int = 2,
                         n_samples_per_class=300) -> Tuple[np.ndarray, np.ndarray]:
    """
    Generate spiral data for classification problem.

    Args:
        data_limit (int, optional): range of data. Defaults to 30.
        n_classes (int, optional): number of classes for classification. Defaults to 2.
        n_samples_per_class (int, optional): number of samples per classes. Defaults to 300.

    Returns:
        Tuple[np.ndarray, np.ndarray]: X,y.
    """

    theta = np.pi * (2 / n_classes)
    rotation_mat = rotation_matrix_2d(theta=theta)

    features = [np.array([
        [
            (np.cos(i/8) * i) + np.random.randn(), (np.sin(i/8) * i) + np.random.randn()
        ] for i in np.linspace(2, data_limit, n_samples_per_class)
    ])]
    target = [np.ones(shape=(n_samples_per_class, 1), dtype=np.int32) * 0]

    for i in range(0, n_classes-1):
        features.append(features[-1] @ rotation_mat)
        target.append(np.ones(shape=(n_samples_per_class, 1), dtype=np.int32) * i+1)
    X = np.vstack(features)
    y = np.vstack(target)
    return X, y

def sine_wave_from_sample(
        n_samples: int, signal_freq: float, n_cycles: int = 10, amplitude: int = 1,
        amp_shift: int = 0, phase_shift: int = 0) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Sine wave generation with number of samples and signal frequency

    Reference:
        https://machinelearningexploration.readthedocs.io/en/latest/MathExploration/Fourier.html#Sine-wave

    Args:
        n_samples (int): number of samples.
        signal_freq (float): signal frequency.
        n_cycles (int, optional): number of cycles. Defaults to 10.
        amplitude (int, optional): signal amplitude. Defaults to 1.
        amp_shift (int, optional): amplitude shift. Defaults to 0.
        phase_shift (int, optional): phase shift. Defaults to 0.

    Returns:
        Tuple[np.ndarray, np.ndarray, np.ndarray]: signal wave, time, freq.
    """
    sampling_freq = (n_samples * signal_freq) / n_cycles
    # n_samples = int((sampling_freq / signal_freq) * n_cycles)

    t_step = 1 / sampling_freq
    time = np.linspace(0, (n_samples-1) * t_step, n_samples)

    f_step = sampling_freq / n_samples
    freq = np.linspace(0, (n_samples-1) * f_step, n_samples)

    wave = amp_shift + amplitude * \
        np.sin((2 * np.pi * signal_freq * time) + phase_shift)

    return wave, time, freq

def sine_wave_from_timesteps(
        signal_freq: float, time_step: float, amplitude: int = 1, amp_shift: int = 0,
        phase_shift: int = 0) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Sine wave generation with time steps and signal frequency

    Args:
        signal_freq (float): singal frequency.
        time_step (float): time step.
        amplitude (int, optional): amplitude. Defaults to 1.
        amp_shift (int, optional): amplitude shift. Defaults to 0.
        phase_shift (int, optional): phase shift. Defaults to 0.

    Returns:
        Tuple[np.ndarray, np.ndarray, np.ndarray]: signal wave, time, freq.
    """
    time = np.arange(0, 1, time_step)
    n = len(time)
    freq = (1/(time_step * n)) * np.arange(n)

    wave = amp_shift + amplitude * \
        np.sin((2 * np.pi * signal_freq * time) + phase_shift)

    return wave, time, freq
        
if __name__ == "__main__":

    # import seaborn as sns
    # import matplotlib.pyplot as plt

    # X, y = generate_spiral_data(n_classes=1)

    # sns.scatterplot(x=X[..., 0], y=X[..., 1], hue=y[..., 0], palette='dark')
    # plt.show()

    pass


import numpy as np
from mightypy.stats import population_stability_index

np.random.seed(10)

expected_continuous = np.random.normal(size=(500,))
actual_continuous = np.random.normal(size=(500,))
psi_df = population_stability_index(expected_continuous, actual_continuous, data_type='continuous')
print(psi_df.psi.sum())


expected_discrete = np.random.randint(0,10, size=(500,))
actual_discrete = np.random.randint(0,10, size=(500,))
psi_df = population_stability_index(expected_discrete, actual_discrete, data_type='discrete')
print(psi_df.psi.sum())


expected_continuous = np.random.normal(size=(500,))
actual_continuous = np.random.normal(size=(500,))
psi_df = population_stability_index(expected_continuous, actual_continuous, data_type='some_random_number')
print(psi_df.psi.sum())


"""
Utiltiy helper functions of Machine Learning
"""

import numpy as np


def sigmoid(val: np.ndarray) -> np.ndarray:
    """Sigmoid function

    .. math::
        f(z) = \\frac{1}{1 + e^{-z}}

    Args:
        val (ndarray): input value

    Returns:
        np.ndarray: sigmoid value
    """
    return 1 / (1 + np.exp(-val))


def moving_window_matrix(arr: np.ndarray, window: int, lag: int = 1) -> np.ndarray:
    """Create Moving Window matrix for 1D data.

    More details on this function.
    https://machinelearningexploration.readthedocs.io/en/latest/MathExploration/MovingWindow.html

    Args:
        arr (np.ndarray): input 1D array.
        window (int): window/ number of columns.
        lag (int, optional): lag count for moving. Defaults to 1.

    Returns:
        np.ndarray: transformed matrix.

    Raises:
        AssertionError: input array shape should be 1D like (m,).
        AssertionError: length of array should be greater than window size and lag.

    Example:
        >>> a = np.random.rand(100)
        >>> print(moving_window_matrix(a, 20, 2))
    """

    assert len(np.shape(arr)) == 1, 'input array shape should be 1D like (m,).'
    size = arr.shape[0]
    assert size > window and size > lag, \
        'length of array should be greater than window size and lag.'

    frame_width = size - window + 1

    new_frame_width = int(np.ceil(frame_width / lag))
    new_frame = np.empty(shape=(window, new_frame_width))
    for row in range(0, window):
        new_frame[row] = arr[row: row+frame_width][::lag]

    return new_frame.T



if __name__ == "__main__":

    # a = np.random.rand(100)

    # print(moving_window_matrix(a, 20, 2))

    # import matplotlib.pyplot as plt

    # x = np.array([1, 2, 3, 4, 5, 6, 7, 8])
    # y = np.array([1, 2, 3, 3, 4, 5, 7, 10])
    # s, r, t = trend(x, y)
    # plt.plot(x, y, 'o', label='original')
    # plt.plot(x, t, '.-',  label='regression line')
    # plt.legend()
    # plt.show()


    # x = np.arange(-10, 10)
    # y = x**2 + x**3
    # s, r, l = polynomial_regression(x, y, 3)

    # plt.plot(x, y, 'ko', label='original')
    # plt.plot(x, l, '.-',  label='regression line')
    # plt.legend()
    # plt.show()

    pass



from typing import Callable, Tuple, Optional
import numpy as np
import pandas as pd


def woe_and_iv(df: pd.DataFrame, event: str, non_event: str, target_col: str, bucket_col: str,
               value_col: Optional[str] = None, agg_func: Callable = np.count_nonzero, bucket_col_type: str = 'continuous',
               n_buckets: int = 10) -> Tuple[pd.DataFrame, float]:
    """
    Weight of Evidence and Information Value.

    References:
    -----------
        https://www.listendata.com/2015/03/weight-of-evidence-woe-and-information.html

    Args:
        df (pd.DataFrame): Input pandas dataframe.
        event (str): event name. Generally label true/1.
        non_event (str): non event name. Generally label false/0.
        target_col (str): Target column name.
        value_col (str): Value column name to aggregate(count). Defaults to None.
        bucket_col (str): bucketing column name.
        agg_func (Callable, optional): Aggregation function name. Defaults to np.count_nonzero.
        bucket_col_type (str, optional): Bucketing columns value type. If discrete buckets will not be created else buckets will be created.
                                            Defaults to 'continuous'.
        n_buckets (int, optional): If bucket column has continuous values then create aritificial buckets. Defaults to 10.

    Raises:
        NotImplementedError: If bucket column type is not 'continuous' or 'discrete'.

    Returns:
        Tuple[pd.DataFrame, float]: calculated dataframe with weight of evidence, Information Value.

    Examples:
        >>> from sklearn.datasets import load_breast_cancer
        >>> from mightypy.stats import woe_and_iv 
        >>> dataset = load_breast_cancer(as_frame=True)
        >>> df = dataset.frame[['mean radius', 'target']]
        >>> target_map = {0: 'False', 1: 'True'}
        >>> df['label'] = df['target'].map(target_map)
        >>> cal_df, iv = woe_and_iv(df, event='True', non_event='False', target_col='label',
        >>>                         bucket_col='mean radius')
    """
    if value_col is None:
        value_col = 'values'
        df.insert(loc=0, column=value_col, value='x')
    df = df[[target_col, value_col, bucket_col]].copy()
    bucket_col_name: str = f'buckets_{bucket_col}'
    perc_event_col_name: str = f'%_event_{event}'
    perc_non_event_col_name: str = f'%_non_event_{non_event}'

    if bucket_col_type == 'continuous':
        quantiles = np.linspace(0, 1, n_buckets+1)
        df.insert(loc=0, column=bucket_col_name,
                  value=pd.qcut(df[bucket_col].values, q=quantiles, duplicates='raise', retbins=False))
    elif bucket_col_type == 'discrete':
        df.insert(loc=0, column=bucket_col_name, value=df[bucket_col])
    else:
        raise NotImplementedError

    cal_df = pd.pivot_table(df, index=[bucket_col_name], columns=[target_col],
                            values=value_col, aggfunc=agg_func)

    cal_df.fillna(value=0, inplace=True)

    cal_df[['adj_event', 'adj_non_event']] = cal_df[[event,non_event]]\
        .apply(lambda x: (x+0.5) if (x[0] == 0 or x[1] == 0) else x, axis=1)

    event_sum = cal_df[event].sum()
    cal_df[perc_event_col_name] = cal_df['adj_event'] / event_sum

    non_event_sum = cal_df[non_event].sum()
    cal_df[perc_non_event_col_name] = cal_df['adj_non_event'] / non_event_sum

    cal_df['woe'] = np.log(
        cal_df[perc_non_event_col_name] / cal_df[perc_event_col_name]
    )

    cal_df['iv'] = (
        cal_df[perc_non_event_col_name] -
        cal_df[perc_event_col_name]
    ) * cal_df['woe']

    iv: float = cal_df['iv'].sum()
    return cal_df, iv


if __name__ == "__main__":
    pass



"""
Linked list setup
"""

class Node:
    def __init__(self, data):
        self.data = data 
        self._next = None

    @property
    def next(self):
        return self._next

    @next.setter
    def next(self,new_next):
        self._next = new_next

    def __repr__(self):
        return f"Node Value : {self.data}"

class LinkedList:
    def __init__(self):
        self.head = None
        self.tail = None

    def append(self,value):
        """Insert at the end.

        Args:
            value (object): Value to insert in linked list
        """
        temp = Node(value)
        if self.head is None:
            self.head = temp
        elif self.tail is None:
            self.tail = temp
            self.head.next = self.tail
        else:
            self.tail.next = temp
            self.tail = self.tail.next


    def push(self,value):
        """Push at the begining.

        Args:
            value (object): Value to push to list
        """
        temp = Node(value)
        if self.head is None:
            self.head = temp
        elif self.tail is None:
            self.tail = self.head
            self.head = temp
            self.head.next = self.tail
        else:
            temp.next = self.head
            self.head = temp

    def traverse(self):
        """Traverse the linked list.

        Returns:
            node_values (list) : linked list values
        """
        node_values = []
        current_node = self.head
        while current_node:
            node_values.append(current_node.data)
            current_node = current_node.next

        return node_values

    def __len__(self):

        size = 0
        curr_node = self.head

        while curr_node:
            size += 1
            curr_node = curr_node.next

        return size


if __name__ == "__main__":
    
    l_list = LinkedList()

    for i in range(10):
        l_list.push("push"+str(i))

    for i in range(10):
        l_list.append("insert"+str(i))


    print(l_list.traverse())

    print(len(l_list))
